package ru.dbotthepony.kstarbound

import com.google.gson.Gson
import com.google.gson.JsonObject
import com.google.gson.internal.bind.JsonTreeReader
import it.unimi.dsi.fastutil.ints.Int2ObjectMap
import it.unimi.dsi.fastutil.ints.Int2ObjectOpenHashMap
import it.unimi.dsi.fastutil.ints.IntCollection
import it.unimi.dsi.fastutil.ints.IntIterator
import it.unimi.dsi.fastutil.ints.IntSet
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectCollection
import it.unimi.dsi.fastutil.objects.ObjectIterator
import it.unimi.dsi.fastutil.objects.ObjectSet
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.api.IStarboundFile
import ru.dbotthepony.kstarbound.util.PathStack
import java.util.*
import kotlin.reflect.KClass

inline fun <reified T : Any> ObjectRegistry(name: String, noinline key: (T) -> String, noinline intKey: ((T) -> Int)? = null): ObjectRegistry<T> {
	return ObjectRegistry(T::class, name, key, intKey)
}

class RegistryObject<T : Any>(val value: T, private val json: JsonObject, val file: IStarboundFile, val gson: Gson, val pathStack: PathStack) {
	fun copy(): JsonObject {
		return json.deepCopy()
	}

	override fun equals(other: Any?): Boolean {
		return other is RegistryObject<*> && other.value == value && other.json == json
	}

	override fun hashCode(): Int {
		return value.hashCode().rotateRight(13) xor json.hashCode()
	}

	override fun toString(): String {
		return "RegistryObject[$value from $file]"
	}
}

class ObjectRegistry<T : Any>(val clazz: KClass<T>, val name: String, val key: (T) -> String, val intKey: ((T) -> Int)? = null) {
	private val objects = Object2ObjectOpenHashMap<String, RegistryObject<T>>()
	private val intObjects = Int2ObjectOpenHashMap<RegistryObject<T>>()

	val view: Map<String, RegistryObject<T>> = Collections.unmodifiableMap(objects)
	val intView = object : Int2ObjectMap<RegistryObject<T>> {
		override fun get(key: Int): RegistryObject<T>? = intObjects[key]
		override fun containsKey(key: Int) = intObjects.containsKey(key)
		override fun defaultReturnValue(rv: RegistryObject<T>) = throw UnsupportedOperationException()
		override fun defaultReturnValue(): RegistryObject<T> = throw UnsupportedOperationException()
		override fun containsValue(value: RegistryObject<T>) = intObjects.containsValue(value)
		override fun isEmpty() = intObjects.isEmpty()
		override fun putAll(from: Map<out Int, RegistryObject<T>>) = throw UnsupportedOperationException()

		private val int2ObjectEntrySet: ObjectSet<Int2ObjectMap.Entry<RegistryObject<T>>> = object : ObjectSet<Int2ObjectMap.Entry<RegistryObject<T>>> {
			override val size: Int
				get() = intObjects.int2ObjectEntrySet().size

			override fun add(element: Int2ObjectMap.Entry<RegistryObject<T>>) = throw UnsupportedOperationException()
			override fun addAll(elements: Collection<Int2ObjectMap.Entry<RegistryObject<T>>>) = throw UnsupportedOperationException()
			override fun clear() = throw UnsupportedOperationException()

			override fun iterator(): ObjectIterator<Int2ObjectMap.Entry<RegistryObject<T>>> {
				return object : ObjectIterator<Int2ObjectMap.Entry<RegistryObject<T>>> {
					val iterator = intObjects.int2ObjectEntrySet().iterator()
					override fun hasNext() = iterator.hasNext()
					override fun remove() = throw UnsupportedOperationException()
					override fun next() = iterator.next()
				}
			}

			override fun contains(element: Int2ObjectMap.Entry<RegistryObject<T>>) = intObjects.int2ObjectEntrySet().contains(element)
			override fun containsAll(elements: Collection<Int2ObjectMap.Entry<RegistryObject<T>>>) = intObjects.int2ObjectEntrySet().containsAll(elements)
			override fun isEmpty() = intObjects.int2ObjectEntrySet().isEmpty()
			override fun remove(element: Int2ObjectMap.Entry<RegistryObject<T>>) = throw UnsupportedOperationException()
			override fun removeAll(elements: Collection<Int2ObjectMap.Entry<RegistryObject<T>>>) = throw UnsupportedOperationException()
			override fun retainAll(elements: Collection<Int2ObjectMap.Entry<RegistryObject<T>>>) = throw UnsupportedOperationException()
		}

		override fun int2ObjectEntrySet(): ObjectSet<Int2ObjectMap.Entry<RegistryObject<T>>> {
			return int2ObjectEntrySet
		}

		override val keys: IntSet = object : IntSet {
			override fun add(key: Int) = throw UnsupportedOperationException()
			override fun addAll(c: IntCollection) = throw UnsupportedOperationException()
			override fun addAll(elements: Collection<Int>) = throw UnsupportedOperationException()
			override fun clear() = throw UnsupportedOperationException()

			override fun iterator(): IntIterator {
				return object : IntIterator {
					val iterator = intObjects.keys.iterator()
					override fun hasNext() = iterator.hasNext()
					override fun remove() = throw UnsupportedOperationException()
					override fun nextInt() = iterator.nextInt()
				}
			}

			override fun remove(k: Int) = throw UnsupportedOperationException()
			override fun removeAll(c: IntCollection) = throw UnsupportedOperationException()
			override fun removeAll(elements: Collection<Int>) = throw UnsupportedOperationException()
			override fun retainAll(c: IntCollection) = throw UnsupportedOperationException()
			override fun retainAll(elements: Collection<Int>) = throw UnsupportedOperationException()
			override fun contains(key: Int) = intObjects.keys.contains(key)
			override fun containsAll(c: IntCollection) = intObjects.keys.containsAll(c)
			override fun containsAll(elements: Collection<Int>) = intObjects.keys.containsAll(elements)
			override fun isEmpty() = intObjects.keys.isEmpty()
			override fun toArray(a: IntArray) = intObjects.keys.toArray(a)
			override fun toIntArray() = intObjects.keys.toIntArray()

			override val size: Int
				get() = intObjects.keys.size
		}

		override val values: ObjectCollection<RegistryObject<T>> = object : ObjectCollection<RegistryObject<T>> {
			override val size: Int
				get() = intObjects.values.size

			override fun add(element: RegistryObject<T>) = throw UnsupportedOperationException()
			override fun addAll(elements: Collection<RegistryObject<T>>) = throw UnsupportedOperationException()
			override fun clear() = throw UnsupportedOperationException()

			override fun iterator(): ObjectIterator<RegistryObject<T>> {
				return object : ObjectIterator<RegistryObject<T>> {
					val iterator = intObjects.values.iterator()
					override fun hasNext() = iterator.hasNext()
					override fun next(): RegistryObject<T> = iterator.next()
					override fun remove() = throw UnsupportedOperationException()
				}
			}

			override fun contains(element: RegistryObject<T>) = intObjects.values.contains(element)
			override fun containsAll(elements: Collection<RegistryObject<T>>) = intObjects.values.containsAll(elements)
			override fun isEmpty() = intObjects.values.isEmpty()
			override fun remove(element: RegistryObject<T>) = throw UnsupportedOperationException()
			override fun removeAll(elements: Collection<RegistryObject<T>>) = throw UnsupportedOperationException()
			override fun retainAll(elements: Collection<RegistryObject<T>>) = throw UnsupportedOperationException()
		}

		override val size: Int
			get() = intObjects.size
	}

	fun clear() {
		objects.clear()
		intObjects.clear()
	}

	fun add(gson: Gson, file: IStarboundFile, pathStack: PathStack): Boolean {
		return pathStack(file.computeDirectory()) {
			val elem = gson.fromJson(file.reader(), JsonObject::class.java)
			val value = gson.fromJson<T>(JsonTreeReader(elem), clazz.java)
			add(RegistryObject(value, elem, file, gson, pathStack))
		}
	}

	fun add(value: T, json: JsonObject, file: IStarboundFile, gson: Gson, pathStack: PathStack): Boolean {
		return add(RegistryObject(value, json, file, gson, pathStack))
	}

	private fun add(value: RegistryObject<T>): Boolean {
		val key = this.key.invoke(value.value)

		val existing = objects.put(key, value)

		if (existing != null) {
			LOGGER.warn("Registry $name already has object with key $key! Overwriting. (old originated from ${existing.file}, new originate from ${value.file}).")
		}

		if (this.intKey == null)
			return existing != null

		val intKey = this.intKey.invoke(value.value)
		val intExisting = intObjects.put(intKey, value)

		if (intExisting != null) {
			LOGGER.warn("Registry $name already has object with ID $intKey (new $key, old ${this.key.invoke(intExisting.value)})! Overwriting. (old originated from ${intExisting.file}, new originate from ${value.file}).")
		}

		return existing != null || intExisting != null
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
