package ru.dbotthepony.kstarbound.client.gl.shader

import ru.dbotthepony.kstarbound.client.gl.BlendFunc
import ru.dbotthepony.kstarbound.client.gl.GLStateTracker
import ru.dbotthepony.kstarbound.client.gl.GLType
import ru.dbotthepony.kstarbound.client.gl.vertex.GLAttributeList
import ru.dbotthepony.kstarbound.client.gl.vertex.StreamVertexBuilder
import ru.dbotthepony.kstarbound.client.gl.vertex.GeometryType
import ru.dbotthepony.kstarbound.client.gl.vertex.QuadTransformers
import ru.dbotthepony.kstarbound.client.render.GPULightRenderer
import ru.dbotthepony.kvector.matrix.nfloat.Matrix4f
import ru.dbotthepony.kvector.vector.Color

private fun GLStateTracker.shaders(name: String): List<GLStateTracker.Shader> {
	return listOf(internalVertex("shaders/$name.vsh"), internalFragment("shaders/$name.fsh"))
}

private fun GLStateTracker.gshaders(name: String): List<GLStateTracker.Shader> {
	return listOf(
		internalVertex(name),
		internalFragment(name),
		internalGeometry(name)
	)
}

class GLLiquidProgram(state: GLStateTracker) : GLShaderProgram(state, state.shaders("liquid")) {
	var baselineColor by F4Uniform("baselineColor")
	var transform by F4x4Uniform("transform")

	val builder by lazy {
		StreamVertexBuilder(state, FORMAT, GeometryType.QUADS, 16384)
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).build()
	}
}

class GLLightProgram(state: GLStateTracker) : GLShaderProgram(state, state.shaders("light")) {
	var baselineColor by F4Uniform("baselineColor")
	var transform by F4x4Uniform("transform")

	val builder by lazy {
		StreamVertexBuilder(state, FORMAT, GeometryType.QUADS, 32)
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).push(GLType.VEC2F).build()
	}
}

class GLHardLightGeometryProgram(state: GLStateTracker) : GLShaderProgram(state, state.gshaders("hard_light_geometry")) {
	var transform by F4x4Uniform("transform")
	var localToWorldTransform by F4x4Uniform("localToWorldTransform")
	var lightPosition by F2Uniform("lightPosition")
	var lightPenetration by FUniform("lightPenetration")

	val builder by lazy {
		StreamVertexBuilder(state, GPULightRenderer.SHADOW_FORMAT, GeometryType.QUADS_AS_LINES, 32)
	}
}

class GLSoftLightGeometryProgram(state: GLStateTracker) : GLShaderProgram(state, state.shaders("soft_light_geometry2")) {
	var transform by F4x4Uniform("transform")
	var lightPenetration by FUniform("lightPenetration")
	var localToWorldTransform by F4x4Uniform("localToWorldTransform")

	/**
	 * Vector3f(x, y, size)
	 */
	var lightPositionAndSize by F3Uniform("lightPositionAndSize")

	val builder by lazy {
		StreamVertexBuilder(state, GPULightRenderer.SHADOW_FORMAT_SOFT, GeometryType.QUADS_AS_LINES, 32)
	}
}

class GLColorQuadProgram(state: GLStateTracker) : GLShaderProgram(state, state.shaders("screen_quad")) {
	var color by F4Uniform("color")

	private val builder by lazy {
		val builder = StreamVertexBuilder(state, FORMAT, GeometryType.QUADS, 1)

		builder.builder.begin()
		builder.builder.quad(-1f, -1f, 1f, 1f)
		builder.upload()

		builder
	}

	fun clearAlpha() {
		use()

		color = ALPHA

		val old = state.blend
		val oldFunc = state.blendFunc

		state.blend = true
		state.blendFunc = BlendFunc.ONLY_ALPHA
		builder.draw()
		state.blend = old
		state.blendFunc = oldFunc
	}

	fun clearColor(color: Color = Color.WHITE) {
		use()

		this.color = color

		val old = state.blend

		state.blend = false
		builder.draw()
		state.blend = old
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).build()
		val ALPHA = Color(0f, 0f, 0f, 1f)
	}
}

class GLTextureQuadProgram(state: GLStateTracker) : GLShaderProgram(state, state.shaders("screen_quad_tex")) {
	var texture by IUniform("texture0")

	private val builder by lazy {
		val builder = StreamVertexBuilder(state, FORMAT, GeometryType.QUADS, 1)

		builder.builder.begin()
		builder.builder.quad(-1f, -1f, 1f, 1f, QuadTransformers.uv())
		builder.upload()

		builder
	}

	fun run(texture: Int = 0) {
		use()
		this.texture = texture
		builder.draw()
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).push(GLType.VEC2F).build()
	}
}

class GLTextureBlurredQuadProgram(state: GLStateTracker) : GLShaderProgram(state, state.shaders("screen_quad_tex_blur")) {
	var texture by IUniform("texture0")

	private val builder by lazy {
		val builder = StreamVertexBuilder(state, FORMAT, GeometryType.QUADS, 1)

		builder.builder.begin()
		builder.builder.quad(-1f, -1f, 1f, 1f, QuadTransformers.uv())
		builder.upload()

		builder
	}

	fun run(texture: Int = 0) {
		use()
		this.texture = texture
		builder.draw()
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).push(GLType.VEC2F).build()
	}
}

class GLFlatColorProgram(state: GLStateTracker) : GLShaderProgram(state, state.shaders("flat_color")) {
	var transform by F4x4Uniform("transform")

	val builder by lazy {
		StreamVertexBuilder(state, FORMAT, GeometryType.QUADS, 16384)
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).push(GLType.VEC4F).build()
	}
}

class GLTileProgram(state: GLStateTracker) : GLShaderProgram(state, state.shaders("tile")) {
	var transform by F4x4Uniform("transform")
	var color by F4Uniform("color")
	var texture by IUniform("texture0")

	init {
		transform = Matrix4f.IDENTITY
		color = Color.WHITE
	}
}

class GLFontProgram(state: GLStateTracker) : GLShaderProgram(state, state.shaders("font")) {
	var transform by F4x4Uniform("transform")
	var color by F4Uniform("color")
	var texture by IUniform("texture0")

	init {
		transform = Matrix4f.IDENTITY
		color = Color.WHITE
	}
}

class GLFlatProgram(state: GLStateTracker) : GLShaderProgram(state, state.shaders("flat")) {
	var transform by F4x4Uniform("transform")
	var color by F4Uniform("color")

	init {
		color = Color.WHITE
	}
}

class GLTexturedProgram(state: GLStateTracker) : GLShaderProgram(state, listOf(state.internalVertex("shaders/vertex/texture.glsl"), state.internalFragment("shaders/fragment/texture.glsl"))) {
	var transform by F4x4Uniform("transform")
	var texture by IUniform("texture0")

	init {
		transform = Matrix4f.IDENTITY
	}
}

class GLTexturedColoredProgram(state: GLStateTracker) : GLShaderProgram(state, listOf(state.internalVertex("shaders/vertex/texture.glsl"), state.internalFragment("shaders/fragment/texture_color.glsl"))) {
	var transform by F4x4Uniform("transform")
	var texture by IUniform("texture0")
	var color by F4Uniform("color")

	init {
		transform = Matrix4f.IDENTITY
		color = Color.WHITE
	}
}

class GLPrograms(val state: GLStateTracker) {
	val tile by lazy { GLTileProgram(state) }
	val font by lazy { GLFontProgram(state) }
	val flat by lazy { GLFlatProgram(state) }
	val flatColor by lazy { GLFlatColorProgram(state) }
	val liquid by lazy { GLLiquidProgram(state) }
	val light by lazy { GLLightProgram(state) }
	val hardLightGeometry by lazy { GLHardLightGeometryProgram(state) }
	val softLightGeometry by lazy { GLSoftLightGeometryProgram(state) }
	val textured by lazy { GLTexturedProgram(state) }
	val texturedColored by lazy { GLTexturedColoredProgram(state) }

	val viewColorQuad by lazy { GLColorQuadProgram(state) }
	val viewTextureQuad by lazy { GLTextureQuadProgram(state) }
	val viewTextureBlurQuad by lazy { GLTextureBlurredQuadProgram(state) }
}
