package ru.dbotthepony.kstarbound.client.gl.vertex

import org.lwjgl.opengl.GL46
import org.lwjgl.opengl.GL46.GL_UNSIGNED_INT
import org.lwjgl.opengl.GL46.GL_UNSIGNED_SHORT
import org.lwjgl.opengl.GL46.GL_UNSIGNED_BYTE
import ru.dbotthepony.kstarbound.client.gl.GLType
import ru.dbotthepony.kstarbound.client.gl.VertexBufferObject
import ru.dbotthepony.kvector.util2d.AABB
import java.nio.ByteBuffer
import java.nio.ByteOrder
import kotlin.math.cos
import kotlin.math.sin

private fun interface IndexWriter {
	fun write(buffer: ByteBuffer, value: Int)
}

private fun writer(type: Int): IndexWriter {
	return when (type) {
		GL_UNSIGNED_SHORT -> IndexWriter { it, value -> it.putShort(value.toShort()) }
		GL_UNSIGNED_BYTE -> IndexWriter { it, value -> it.put(value.toByte()) }
		else -> IndexWriter { it, value -> it.putInt(value) }
	}
}

private fun chooseIndexType(capacity: Int): Int {
	return when (capacity) {
		// api performance issue 102: glDrawElements uses element index type 'GL_UNSIGNED_BYTE' that is not optimal for the current hardware configuration; consider using 'GL_UNSIGNED_SHORT' instead
		// in 0 .. 255 -> GL_UNSIGNED_BYTE
		in 0 .. 65535 -> GL_UNSIGNED_SHORT
		else -> GL_UNSIGNED_INT
	}
}

private fun indexSize(type: Int): Int {
	return when (type) {
		GL_UNSIGNED_BYTE -> 1
		GL_UNSIGNED_SHORT -> 2
		GL_UNSIGNED_INT -> 4
		else -> throw IllegalStateException()
	}
}

/**
 * Создаёт буферы вне кучи для загрузки геометрии в них, с набором аттрибутов [attributes], типом [type] и указанным начальным размером.
 *
 * По мере необходимости, буферы могут быть увеличены в размерах.
 *
 * Загрузка в память видеокарты происходит напрямую из буферов, через метод [upload]
 */
@Suppress("unchecked_cast")
class VertexBuilder(
	val attributes: GLAttributeList,
	val type: GeometryType,
	initialCapacity: Int = 64
) {
	/**
	 * [GL_UNSIGNED_BYTE], [GL_UNSIGNED_SHORT] или [GL_UNSIGNED_INT]
	 */
	var indexType: Int = chooseIndexType(initialCapacity)
		private set
	var indexSize: Int = indexSize(indexType)
		private set

	/**
	 * В элементах, не вершинах
	 */
	private var capacity = initialCapacity
	private var vertexMemory = ByteBuffer.allocateDirect(capacity * attributes.stride * type.elements).also { it.order(ByteOrder.LITTLE_ENDIAN) }
	private var indexMemory = ByteBuffer.allocateDirect(capacity * indexSize * type.indices.size).also { it.order(ByteOrder.LITTLE_ENDIAN) }
	private var indexWriter = writer(indexType)

	private var inVertex = false
	private var attributeIndex = 0
	private var elementIndexOffset = 0

	private var elementVertices = 0

	var vertexCount = 0
		private set

	var indexCount = 0
		private set

	var elementCount = 0
		private set

	fun begin() {
		inVertex = false
		attributeIndex = 0
		elementIndexOffset = 0
		vertexCount = 0
		indexCount = 0
		elementCount = 0
		vertexMemory.position(0)
		indexMemory.position(0)
	}

	private fun checkBounds() {
		if (attributeIndex >= attributes.size) {
			throw IndexOutOfBoundsException("Tried to add new attribute when already added all attributes")
		}
	}

	fun expect(type: GLType) {
		checkBounds()

		if (attributes[attributeIndex].glType != type) {
			throw IllegalStateException("Expected attribute type $type, got ${attributes[attributeIndex].name}[${attributes[attributeIndex].glType}] (at position $attributeIndex)")
		}
	}

	fun expect(name: String) {
		checkBounds()

		if (attributes[attributeIndex].name != name) {
			throw IllegalStateException("Expected attribute name $name, got ${attributes[attributeIndex].name}[${attributes[attributeIndex].glType}] (at position $attributeIndex)")
		}
	}

	fun upload(vbo: VertexBufferObject, ebo: VertexBufferObject, drawType: Int = GL46.GL_DYNAMIC_DRAW) {
		require(vbo.isArray) { "$vbo is not an array" }
		require(ebo.isElementArray) { "$ebo is not an element array" }

		end()

		check(elementVertices == 0) { "Not fully built vertex element ($type requires ${type.elements} vertex points to be present, yet last strip has only $elementVertices elements)" }

		val vertexPos = vertexMemory.position()
		val elementPos = indexMemory.position()

		vertexMemory.position(0)
		indexMemory.position(0)

		vbo.bufferData(vertexMemory, drawType, length = vertexPos.toLong())
		ebo.bufferData(indexMemory, drawType, length = elementPos.toLong())

		vertexMemory.position(vertexPos)
		indexMemory.position(elementPos)
	}

	fun end() {
		if (inVertex) {
			inVertex = false

			if (attributeIndex != attributes.size) {
				throw IllegalStateException("Unfinished vertex, we are at $attributeIndex, while we have ${attributes.size} attributes")
			}

			vertexCount++
			elementVertices++

			if (elementVertices == type.elements) {
				elementCount++
				elementVertices = 0

				for (index in type.indices) {
					indexWriter.write(indexMemory, index + elementIndexOffset)
				}

				elementIndexOffset += type.elements
				indexCount += type.indices.size

				if (capacity <= elementCount) {
					check(this.vertexMemory.position() == this.vertexMemory.capacity()) { "${this.vertexMemory.position()} != ${this.vertexMemory.capacity()}" }
					check(this.indexMemory.position() == this.indexMemory.capacity()) { "${this.indexMemory.position()} != ${this.indexMemory.capacity()}" }

					val capacity = capacity * 2
					val indexType = chooseIndexType(capacity)
					val indexSize = indexSize(indexType)

					val vertexMemory = ByteBuffer.allocateDirect(capacity * attributes.stride * type.elements)
					vertexMemory.order(ByteOrder.LITTLE_ENDIAN)

					this.vertexMemory.position(0)
					vertexMemory.put(this.vertexMemory)

					val indexMemory = ByteBuffer.allocateDirect(capacity * indexSize * type.indices.size)
					indexMemory.order(ByteOrder.LITTLE_ENDIAN)

					if (indexType != this.indexType) {
						when (this.indexType) {
							GL_UNSIGNED_BYTE -> {
								when (indexType) {
									GL_UNSIGNED_BYTE -> throw IllegalArgumentException()
									GL_UNSIGNED_SHORT -> {
										this.indexMemory.position(0)

										for (i in 0 until this.capacity) {
											indexMemory.putShort(this.indexMemory.get().toShort())
										}
									}
									GL_UNSIGNED_INT -> {
										this.indexMemory.position(0)

										for (i in 0 until this.capacity) {
											indexMemory.putInt(this.indexMemory.get().toInt())
										}
									}
								}
							}

							GL_UNSIGNED_SHORT -> {
								when (indexType) {
									GL_UNSIGNED_BYTE -> throw IllegalArgumentException()
									GL_UNSIGNED_SHORT -> throw IllegalArgumentException()
									GL_UNSIGNED_INT -> {
										this.indexMemory.position(0)

										for (i in 0 until this.capacity) {
											indexMemory.putInt(this.indexMemory.getShort().toInt())
										}
									}
								}
							}

							else -> throw IllegalArgumentException()
						}
					} else {
						this.indexMemory.position(0)
						indexMemory.put(this.indexMemory)
					}

					this.capacity = capacity
					this.vertexMemory = vertexMemory
					this.indexMemory = indexMemory
					this.indexType = indexType
					this.indexSize = indexSize
					this.indexWriter = writer(indexType)
				}
			}
		}
	}

	fun vertex(): VertexBuilder {
		end()

		inVertex = true
		attributeIndex = 0
		return this
	}

	fun pushVec4f(x: Float, y: Float, z: Float, w: Float): VertexBuilder {
		expect(GLType.VEC4F)
		val memory = vertexMemory
		memory.putFloat(x)
		memory.putFloat(y)
		memory.putFloat(z)
		memory.putFloat(w)
		attributeIndex++
		return this
	}

	fun pushVec3f(x: Float, y: Float, z: Float): VertexBuilder {
		expect(GLType.VEC3F)
		val memory = vertexMemory
		memory.putFloat(x)
		memory.putFloat(y)
		memory.putFloat(z)
		attributeIndex++
		return this
	}

	fun pushVec2f(x: Float, y: Float): VertexBuilder {
		expect(GLType.VEC2F)
		val memory = vertexMemory
		memory.putFloat(x)
		memory.putFloat(y)
		attributeIndex++
		return this
	}

	fun push(value: Float): VertexBuilder {
		expect(GLType.FLOAT)
		vertexMemory.putFloat(value)
		attributeIndex++
		return this
	}

	fun shadowLine(x0: Float, y0: Float, x1: Float, y1: Float): VertexBuilder {
		vertex()
		pushVec4f(x0, y0, x1, y1)
		pushVec2f(0f, 0f)

		vertex()
		pushVec4f(x0, y0, x1, y1)
		pushVec2f(0f, 1f)

		vertex()
		pushVec4f(x0, y0, x1, y1)
		pushVec2f(1f, 1f)

		vertex()
		pushVec4f(x0, y0, x1, y1)
		pushVec2f(1f, 0f)

		return this
	}

	fun dfShadowLine(x0: Float, y0: Float, x1: Float, y1: Float): VertexBuilder {
		shadowLine(x0, y0, x1, y1)
		shadowLine(x1, y1, x0, y0)
		return this
	}

	fun shadowQuad(x0: Float, y0: Float, x1: Float, y1: Float): VertexBuilder {
		shadowLine(x0, y0, x1, y0)
		shadowLine(x1, y0, x1, y1)
		shadowLine(x1, y1, x0, y1)
		shadowLine(x0, y1, x0, y0)
		return this
	}

	// Помощники
	fun quad(
		x0: Float,
		y0: Float,
		x1: Float,
		y1: Float,
		lambda: QuadVertexTransformer = EMPTY_VERTEX_TRANSFORM
	): VertexBuilder {
		check(type.elements == 4) { "Currently building $type" }

		lambda(vertex().pushVec2f(x0, y0), 0).end()
		lambda(vertex().pushVec2f(x1, y0), 1).end()
		lambda(vertex().pushVec2f(x0, y1), 2).end()
		lambda(vertex().pushVec2f(x1, y1), 3).end()

		return this
	}

	fun quadRotated(
		x0: Float,
		y0: Float,
		x1: Float,
		y1: Float,
		x: Float,
		y: Float,
		angle: Double,
		lambda: QuadVertexTransformer = EMPTY_VERTEX_TRANSFORM
	): VertexBuilder {
		check(type.elements == 4) { "Currently building $type" }

		val s = sin(angle).toFloat()
		val c = cos(angle).toFloat()

		lambda(vertex().pushVec2f(x + x0 * c - s * y0, y + s * x0 + c * y0), 0).end()
		lambda(vertex().pushVec2f(x + x1 * c - s * y0, y + s * x1 + c * y0), 1).end()
		lambda(vertex().pushVec2f(x + x0 * c - s * y1, y + s * x0 + c * y1), 2).end()
		lambda(vertex().pushVec2f(x + x1 * c - s * y1, y + s * x1 + c * y1), 3).end()

		return this
	}

	fun quad(aabb: AABB, lambda: QuadVertexTransformer = EMPTY_VERTEX_TRANSFORM): VertexBuilder {
		return quad(
			aabb.mins.x.toFloat(),
			aabb.mins.y.toFloat(),
			aabb.maxs.x.toFloat(),
			aabb.maxs.y.toFloat(),
			lambda
		)
	}

	fun quadZ(
		x0: Float,
		y0: Float,
		x1: Float,
		y1: Float,
		z: Float,
		lambda: QuadVertexTransformer = EMPTY_VERTEX_TRANSFORM
	): VertexBuilder {
		check(type.elements == 4) { "Currently building $type" }

		lambda(vertex().pushVec3f(x0, y0, z), 0).end()
		lambda(vertex().pushVec3f(x1, y0, z), 1).end()
		lambda(vertex().pushVec3f(x0, y1, z), 2).end()
		lambda(vertex().pushVec3f(x1, y1, z), 3).end()

		return this
	}

	fun quadRotatedZ(
		x0: Float,
		y0: Float,
		x1: Float,
		y1: Float,
		z: Float,
		x: Float,
		y: Float,
		angle: Double,
		lambda: QuadVertexTransformer = EMPTY_VERTEX_TRANSFORM
	): VertexBuilder {
		check(type.elements == 4) { "Currently building $type" }

		val s = sin(angle).toFloat()
		val c = cos(angle).toFloat()

		lambda(vertex().pushVec3f(x + x0 * c - s * y0, y + s * x0 + c * y0, z), 0).end()
		lambda(vertex().pushVec3f(x + x1 * c - s * y0, y + s * x1 + c * y0, z), 1).end()
		lambda(vertex().pushVec3f(x + x0 * c - s * y1, y + s * x0 + c * y1, z), 2).end()
		lambda(vertex().pushVec3f(x + x1 * c - s * y1, y + s * x1 + c * y1, z), 3).end()

		return this
	}
}
