package ru.dbotthepony.kstarbound.defs

import ru.dbotthepony.kstarbound.io.json.builder.JsonIgnoreProperty
import ru.dbotthepony.kstarbound.util.INotNullDelegate
import kotlin.properties.ReadWriteProperty
import kotlin.reflect.KProperty

abstract class FreezableDefintionBuilder {
	interface IProperty {
		fun checkAndThrow()
	}

	inner class Nullable<V>(private var value: V? = null) : ReadWriteProperty<Any?, V?>, IProperty {
		override fun getValue(thisRef: Any?, property: KProperty<*>): V? {
			return value
		}

		override fun setValue(thisRef: Any?, property: KProperty<*>, value: V?) {
			if (isFrozen) throw IllegalStateException("$thisRef is frozen!")
			this.value = value
		}

		override fun checkAndThrow() {
			// no op
		}

		override fun equals(other: Any?): Boolean {
			return other is Nullable<*> && other.value == value
		}

		override fun hashCode(): Int {
			return value?.hashCode() ?: 0
		}

		override fun toString(): String {
			return "Nullable[$value]"
		}
	}

	inner class NotNull<V : Any>(private var value: V? = null) : ReadWriteProperty<Any?, V>, IProperty, INotNullDelegate {
		override val isPresent: Boolean
			get() = value != null

		fun getNullable(): V? {
			return value
		}

		override fun getValue(thisRef: Any?, property: KProperty<*>): V {
			return checkNotNull(value) {"${property.name} was not initialized"}
		}

		override fun setValue(thisRef: Any?, property: KProperty<*>, value: V) {
			if (isFrozen) throw IllegalStateException("$thisRef is frozen!")
			this.value = value
		}

		override fun checkAndThrow() {
			if (value == null) {
				throw NullPointerException("Value was not initialized")
			}
		}

		override fun equals(other: Any?): Boolean {
			return other is NotNull<*> && other.value == value
		}

		override fun hashCode(): Int {
			return value?.hashCode() ?: 0
		}

		override fun toString(): String {
			return "NotNull[$value]"
		}
	}

	@JsonIgnoreProperty
	var isFrozen = false
		private set

	@JsonIgnoreProperty
	private val properties = ArrayList<IProperty>()

	override fun equals(other: Any?): Boolean {
		if (other is FreezableDefintionBuilder && other.javaClass === this.javaClass)
			return other.properties == properties

		return false
	}

	override fun hashCode(): Int {
		return properties.hashCode()
	}

	override fun toString(): String {
		return "${this::class.simpleName}[frozen=$isFrozen,values={${properties.joinToString(",")}}]"
	}

	protected open fun onFreeze() {

	}

	fun freeze() {
		if (isFrozen) return
		onFreeze()
		properties.forEach { it.checkAndThrow() }
		isFrozen = true
	}
}
