package ru.dbotthepony.kstarbound.defs

import com.google.common.collect.Interners
import com.google.gson.Gson
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.TypeAdapterFactory
import com.google.gson.reflect.TypeToken
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import it.unimi.dsi.fastutil.objects.Reference2ObjectArrayMap
import ru.dbotthepony.kstarbound.RegistryObject
import java.lang.ref.Reference
import java.lang.ref.ReferenceQueue
import java.lang.ref.WeakReference
import java.lang.reflect.ParameterizedType
import java.util.concurrent.ConcurrentHashMap
import java.util.function.Supplier

class RegistryReferenceFactory : TypeAdapterFactory {
	private val types = Reference2ObjectArrayMap<Class<*>, (String) -> RegistryObject<Nothing>?>()
	private var isLenient = false

	fun lenient(): RegistryReferenceFactory {
		isLenient = true
		return this
	}

	fun <T : Any> add(clazz: Class<T>, resolver: (String) -> RegistryObject<T>?): RegistryReferenceFactory {
		check(types.put(clazz, resolver as (String) -> RegistryObject<Nothing>?) == null) { "Already has resolver for class $clazz!" }
		return this
	}

	inline fun <reified T: Any> add(noinline resolver: (String) -> RegistryObject<T>?) = add(T::class.java, resolver)

	override fun <T : Any?> create(gson: Gson, type: TypeToken<T>): TypeAdapter<T>? {
		if (type.rawType == RegistryReference::class.java) {
			val ptype = type.type as? ParameterizedType ?: return null
			val registryType = ptype.actualTypeArguments[0]
			val resolver = types[registryType] ?: return if (isLenient) null else throw NoSuchElementException("Can't deserialize registry reference with type $registryType!")
			return RegistryReferenceTypeAdapter(resolver, gson.getAdapter(String::class.java)) as TypeAdapter<T>
		}

		return null
	}
}

class RegistryReferenceTypeAdapter<T : Any>(val resolver: (String) -> RegistryObject<T>?, val strings: TypeAdapter<String>) : TypeAdapter<RegistryReference<T>>() {
	override fun write(out: JsonWriter, value: RegistryReference<T>?) {
		if (value == null)
			out.nullValue()
		else
			strings.write(out, value.name)
	}

	override fun read(`in`: JsonReader): RegistryReference<T>? {
		if (`in`.peek() == JsonToken.NULL)
			return null

		if (`in`.peek() == JsonToken.STRING) {
			return RegistryReference(strings.read(`in`)!!, resolver)
		}

		throw JsonSyntaxException("Expecting string for registry reference, ${`in`.peek()} given, near ${`in`.path}")
	}
}

data class RegistryReference<T : Any>(val name: String, val resolver: (String) -> RegistryObject<T>?) : Supplier<RegistryObject<T>?>, () -> RegistryObject<T>?, Lazy<RegistryObject<T>?> {
	private val lazy = lazy { resolver.invoke(name) }

	override fun get(): RegistryObject<T>? {
		return lazy.value
	}

	override val value: RegistryObject<T>?
		get() = lazy.value

	override fun isInitialized(): Boolean {
		return lazy.isInitialized()
	}

	override fun invoke(): RegistryObject<T>? {
		return lazy.value
	}
}
