package ru.dbotthepony.kstarbound.defs.image

import com.google.gson.Gson
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.TypeAdapterFactory
import com.google.gson.reflect.TypeToken
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kstarbound.defs.DirectAssetReference
import ru.dbotthepony.kstarbound.util.PathStack
import ru.dbotthepony.kstarbound.util.SBPattern

/**
 * @see [AtlasConfiguration.Registry.get]
 */
class ImageReference private constructor(
	val raw: DirectAssetReference,
	val imagePath: SBPattern,
	val spritePath: SBPattern?,
	val atlas: AtlasConfiguration?,
	private val atlasLocator: (String) -> AtlasConfiguration?
) {
	/**
	 * Спрайт, на которое ссылается данный референс, или `null` если:
	 * * [atlas] равен `null`
	 * * [spritePath] является шаблоном и определены не все значения
	 * * [spritePath] не является правильным именем спрайта внутри [atlas] (смотрим [AtlasConfiguration.get])
	 */
	val sprite by lazy(LazyThreadSafetyMode.NONE) {
		if (atlas == null)
			null
		else if (spritePath == null)
			atlas.any()
		else
			atlas.get(spritePath.value ?: return@lazy null)
	}

	fun with(values: (String) -> String?): ImageReference {
		val imagePath = this.imagePath.with(values)
		var spritePath = this.spritePath?.with(values)

		if (spritePath == null) {
			val frame = values.invoke("frame")

			if (frame != null) {
				spritePath = SBPattern.FRAME.with(values)
			}
		}

		if (imagePath != this.imagePath || spritePath != this.spritePath) {
			if (imagePath != this.imagePath) {
				val resolved = imagePath.value

				if (resolved == null)
					return ImageReference(raw, imagePath, spritePath, null, atlasLocator)
				else
					return ImageReference(raw, imagePath, spritePath, atlasLocator.invoke(resolved), atlasLocator)
			} else {
				return ImageReference(raw, imagePath, spritePath, atlas, atlasLocator)
			}
		}

		return this
	}

	fun with(values: Map<String, String>): ImageReference {
		return with(values::get)
	}

	override fun equals(other: Any?): Boolean {
		return other is ImageReference && other.imagePath == imagePath && other.spritePath == spritePath && other.atlas == atlas
	}

	override fun hashCode(): Int {
		return imagePath.hashCode() * 31 + spritePath.hashCode()
	}

	override fun toString(): String {
		return "ImageReference[$imagePath:$spritePath]"
	}

	class Factory(private val atlasLocator: (String) -> AtlasConfiguration, private val remapper: PathStack) : TypeAdapterFactory {
		override fun <T : Any?> create(gson: Gson, type: TypeToken<T>): TypeAdapter<T>? {
			if (type.rawType == ImageReference::class.java) {
				return object : TypeAdapter<ImageReference>() {
					private val strings = gson.getAdapter(String::class.java)

					override fun write(out: JsonWriter, value: ImageReference?) {
						out.value(value?.raw?.fullPath)
					}

					override fun read(`in`: JsonReader): ImageReference? {
						if (`in`.peek() == JsonToken.NULL)
							return null

						val path = strings.read(`in`)

						if (path == "")
							return NEVER

						val split = path.split(':')

						if (split.size > 2) {
							throw JsonSyntaxException("Ambiguous image reference: $path")
						}

						val imagePath = if (split.size == 2) SBPattern.of(split[0]) else SBPattern.of(path)
						val spritePath = if (split.size == 2) SBPattern.of(split[1]) else null

						if (imagePath.isPlainString) {
							val remapped = remapper.remap(split[0])
							return ImageReference(DirectAssetReference(path, remapper.remap(path)), SBPattern.raw(remapped), spritePath, atlasLocator.invoke(remapped), atlasLocator)
						} else {
							return ImageReference(DirectAssetReference(path, path), imagePath, spritePath, null, atlasLocator)
						}
					}
				} as TypeAdapter<T>
			}

			return null
		}
	}

	companion object {
		val NEVER = ImageReference(DirectAssetReference("", ""), SBPattern.EMPTY, null, null) { null }
	}
}
