package ru.dbotthepony.kstarbound.io.json.builder

import com.google.gson.Gson
import com.google.gson.TypeAdapter
import com.google.gson.TypeAdapterFactory
import com.google.gson.reflect.TypeToken
import kotlin.reflect.KClass

private fun Int.toBool() = if (this == 0) null else this > 0

/**
 * Указывает, что для данного класса можно автоматически создать [BuilderAdapter] для всех его свойств,
 * которые не указаны как [JsonIgnoreProperty]
 *
 * @see JsonIgnoreProperty
 * @see JsonPropertyConfig
 */
@Target(AnnotationTarget.CLASS)
@Retention(AnnotationRetention.RUNTIME)
annotation class JsonBuilder(
	/**
	 * @see BuilderAdapter.Builder.extraPropertiesAreFatal
	 */
	val extraPropertiesAreFatal: Boolean = false,

	/**
	 * @see BuilderAdapter.Builder.ignoreKey
	 */
	val ignoreKeys: Array<String> = [],

	/**
	 * 0 = null
	 * -1 = false
	 * 1 = true
	 *
	 * @see BuilderAdapter.Builder.logMisses
	 */
	val logMisses: Int = 0,

	/**
	 * Включать ли свойства родительского класса в данный [BuilderAdapter]
	 */
	val includeSuperclassProperties: Boolean = true,
)

val JsonBuilder.realLogMisses get() = logMisses.toBool()

/**
 * Заставляет указанное свойство быть проигнорированным при автоматическом создании [BuilderAdapter]
 */
@Target(AnnotationTarget.PROPERTY)
@Retention(AnnotationRetention.RUNTIME)
annotation class JsonIgnoreProperty

/**
 * Выставляет флаги данному свойству при автоматическом создании [BuilderAdapter]
 *
 * @see BuilderAdapter.Builder.add
 */
@Target(AnnotationTarget.PROPERTY)
@Retention(AnnotationRetention.RUNTIME)
annotation class JsonPropertyConfig(
	val isFlat: Boolean = false,

	val write: Boolean = true,

	val mustBePresent: Int = 0,
)

val JsonPropertyConfig.realMustBePresent get() = mustBePresent.toBool()

/**
 * Указывает, что для данного класса можно автоматически создать [FactoryAdapter]
 *
 * В подавляющем большинстве случаев это работает исключительно с data классами
 *
 * С технической точки зрения, у класса определяются все свойства и происходит поиск *главного* конструктора, а
 * затем аргументы конструктора отражаются на свойства класса
 *
 * @see JsonIgnoreProperty
 * @see JsonPropertyConfig
 */
@Target(AnnotationTarget.CLASS)
@Retention(AnnotationRetention.RUNTIME)
annotation class JsonFactory(
	/**
	 * @see FactoryAdapter.Builder.storesJson
	 */
	val storesJson: Boolean = false,

	/**
	 * @see FactoryAdapter.Builder.logMisses
	 */
	val logMisses: Boolean = true,

	/**
	 * @see FactoryAdapter.Builder.inputAsList
	 * @see FactoryAdapter.Builder.inputAsMap
	 */
	val asList: Boolean = false,
)

/**
 * Позволяет указать, какую реализацию использовать при попытке разобрать JSON структуру
 * с аннотированным типом
 *
 * Пример:
 * ```kotlin
 * @JsonFactory
 * data class Item(val thingDef: IThing, val price: Long)
 *
 * @JsonImplementation(Thing::class)
 * interface IThing { val prop: String }
 *
 * data class Thing(override val prop: String) : IThing
 * ```
 */
@Target(AnnotationTarget.CLASS)
@Retention(AnnotationRetention.RUNTIME)
annotation class JsonImplementation(val implementingClass: KClass<*>)

object JsonImplementationTypeFactory : TypeAdapterFactory {
	override fun <T : Any?> create(gson: Gson, type: TypeToken<T>): TypeAdapter<T>? {
		val delegate = type.rawType.getAnnotation(JsonImplementation::class.java)

		if (delegate != null) {
			return gson.getAdapter(delegate.implementingClass.java) as TypeAdapter<T>?
		}

		return null
	}
}
