package ru.dbotthepony.kstarbound.io.json.builder

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableMap
import com.google.common.collect.Interner
import com.google.gson.Gson
import com.google.gson.JsonArray
import com.google.gson.JsonObject
import com.google.gson.JsonParseException
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.TypeAdapterFactory
import com.google.gson.internal.bind.JsonTreeReader
import com.google.gson.internal.bind.TypeAdapters
import com.google.gson.reflect.TypeToken
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import it.unimi.dsi.fastutil.objects.Object2IntArrayMap
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.util.enrollList
import ru.dbotthepony.kstarbound.defs.util.enrollMap
import ru.dbotthepony.kstarbound.defs.util.flattenJsonElement
import ru.dbotthepony.kstarbound.getValue
import ru.dbotthepony.kstarbound.io.json.ifString
import ru.dbotthepony.kstarbound.setValue
import java.lang.reflect.Constructor
import kotlin.jvm.internal.DefaultConstructorMarker
import kotlin.properties.Delegates
import kotlin.reflect.*
import kotlin.reflect.full.declaredMembers
import kotlin.reflect.full.isSubclassOf
import kotlin.reflect.full.isSupertypeOf
import kotlin.reflect.full.primaryConstructor

/**
 * [TypeAdapter] для классов, которые имеют все свои свойства в главном конструкторе.
 */
class FactoryAdapter<T : Any> private constructor(
	val bound: KClass<T>,
	val types: ImmutableList<IResolvedProperty<T, *>>,
	val asJsonArray: Boolean,
	val storesJson: Boolean,
	val logMisses: Boolean,
	val stringInterner: Interner<String>
) : TypeAdapter<T>() {
	private val name2index = Object2IntArrayMap<String>()
	private val loggedMisses = ObjectArraySet<String>()

	init {
		name2index.defaultReturnValue(-1)

		for ((i, pair) in types.withIndex()) {
			name2index[pair.property.name] = i
		}
	}

	/**
	 * Обычный конструктор класса (без флагов "значения по умолчанию")
	 */
	private val regularFactory: KFunction<T> = bound.constructors.firstOrNull first@{
		var requiredSize = types.size

		if (storesJson)
			requiredSize++

		if (it.parameters.size == requiredSize) {
			val iterator = types.iterator()
			val factoryIterator = it.parameters.iterator()

			while (factoryIterator.hasNext() && iterator.hasNext()) {
				val param = factoryIterator.next()
				val nextParam = iterator.next()

				val a = param.type
				val b = nextParam.type

				if (!a.isSupertypeOf(b) || a.isMarkedNullable != b.isMarkedNullable) {
					return@first false
				}
			}

			if (storesJson) {
				val nextParam = factoryIterator.next()

				if (asJsonArray) {
					if (!(nextParam.type.classifier as KClass<*>).isSubclassOf(List::class)) {
						return@first false
					}
				} else {
					if (!(nextParam.type.classifier as KClass<*>).isSubclassOf(Map::class)) {
						return@first false
					}
				}
			}

			return@first true
		}

		return@first false
	} ?: throw NoSuchElementException("Unable to determine constructor for ${bound.qualifiedName} matching (${types.joinToString(", ")})")

	/**
	 * Синтетический конструктор класса, который создаётся Kotlin'ном, для создания классов со значениями по умолчанию
	 */
	private val syntheticFactory: Constructor<T>? = try {
		val typelist = types.map { (it.type.classifier as KClass<*>).java }.toMutableList()

		if (storesJson)
			if (asJsonArray)
				typelist.add(List::class.java)
			else
				typelist.add(Map::class.java)

		for (i in 0 until (if (types.size % 31 == 0) types.size / 31 else types.size / 31 + 1))
			typelist.add(Int::class.java)

		typelist.add(DefaultConstructorMarker::class.java)

		bound.java.getDeclaredConstructor(*typelist.toTypedArray())
	} catch(_: NoSuchMethodException) {
		null
	}

	private val syntheticPrimitives: Array<Any?>?

	init {
		if (syntheticFactory == null) {
			syntheticPrimitives = null
		} else {
			syntheticPrimitives = arrayOfNulls(syntheticFactory.parameters.size)

			for ((i, param) in syntheticFactory.parameters.withIndex()) {
				val type = param.parameterizedType as? Class<*> ?: continue

				if (type.isPrimitive) {
					syntheticPrimitives[i] = when (type) {
						java.lang.Boolean.TYPE -> false
						java.lang.Integer.TYPE -> 0
						java.lang.Long.TYPE -> 0L
						java.lang.Short.TYPE -> (0).toShort()
						java.lang.Byte.TYPE -> (0).toByte()
						java.lang.Character.TYPE -> (0).toChar()
						java.lang.Float.TYPE -> 0f
						java.lang.Double.TYPE -> 0.0
						else -> throw IllegalArgumentException("mamma mia: ${type.name}")
					}
				}
			}
		}
	}

	override fun write(out: JsonWriter, value: T) {
		out.beginObject()

		for ((field, adapter) in types) {
			out.name(field.name)
			@Suppress("unchecked_cast")
			(adapter as TypeAdapter<Any>).write(out, (field as KProperty1<T, Any>).get(value))
		}

		out.endObject()
	}

	override fun read(reader: JsonReader): T {
		// таблица присутствия значений (если значение true то на i было значение внутри json)
		val presentValues = BooleanArray(types.size + (if (storesJson) 1 else 0))
		val readValues = arrayOfNulls<Any>(types.size + (if (storesJson) 1 else 0))

		if (storesJson)
			presentValues[presentValues.size - 1] = true

		@Suppress("name_shadowing")
		var reader = reader

		// Если нам необходимо читать объект как набор данных массива, то давай
		if (asJsonArray) {
			val iterator = types.iterator()
			var fieldId = 0

			if (storesJson) {
				val readArray = TypeAdapters.JSON_ELEMENT.read(reader)

				if (readArray !is JsonArray) {
					throw JsonParseException("Expected JSON element to be an Array, ${readArray::class.qualifiedName} given")
				}

				reader = JsonTreeReader(readArray)
				readValues[readValues.size - 1] = enrollList(flattenJsonElement(readArray) as List<Any>, stringInterner::intern)
			}

			reader.beginArray()

			while (reader.peek() != JsonToken.END_ARRAY) {
				if (!iterator.hasNext()) {
					val name = fieldId.toString()

					if (!storesJson && loggedMisses.add(name)) {
						LOGGER.warn("${bound.qualifiedName} has no property for storing $name")
					}

					reader.skipValue()
					fieldId++

					continue
				}

				val tuple = iterator.next()
				val (field, adapter) = tuple

				try {
					readValues[fieldId] = adapter.read(reader)
					presentValues[fieldId] = true
				} catch(err: Throwable) {
					throw JsonSyntaxException("Reading field \"${field.name}\" near ${reader.path} for ${bound.qualifiedName}", err)
				}

				fieldId++
			}
		// иначе - читаем как json object
		} else {
			var json: JsonObject by Delegates.notNull()
			val hasFlatValues = types.any { it.isFlat }

			if (storesJson || hasFlatValues) {
				val readMap = TypeAdapters.JSON_ELEMENT.read(reader)

				if (readMap !is JsonObject) {
					throw JsonParseException("Expected JSON element to be a Map, ${readMap::class.qualifiedName} given")
				}

				json = readMap
				reader = JsonTreeReader(readMap)
				readValues[readValues.size - 1] = enrollMap(flattenJsonElement(readMap) as Map<String, Any>, stringInterner::intern)
			}

			reader.beginObject()

			while (reader.peek() != JsonToken.END_OBJECT) {
				val name = reader.nextName()
				val fieldId = name2index.getInt(name)

				if (fieldId == -1) {
					if (!storesJson && !hasFlatValues && logMisses && loggedMisses.add(name)) {
						LOGGER.warn("${bound.qualifiedName} has no property for storing $name")
					}

					reader.skipValue()
				} else {
					val tuple = types[fieldId]

					if (tuple.isFlat) {
						reader.skipValue()
						continue
					}

					val (field, adapter) = tuple

					try {
						readValues[fieldId] = adapter.read(reader)
						presentValues[fieldId] = true
					} catch(err: Throwable) {
						throw JsonSyntaxException("Reading field \"${field.name}\" near ${reader.path} for ${bound.qualifiedName}", err)
					}
				}
			}

			for ((i, property) in types.withIndex()) {
				if (property.isFlat) {
					try {
						val read = property.adapter.read(JsonTreeReader(json))

						if (read != null) {
							presentValues[i] = true
							readValues[i] = read
						}
					} catch(err: Throwable) {
						throw JsonSyntaxException("Reading flat field \"${property.property.name}\" near ${reader.path} for ${bound.qualifiedName}", err)
					}
				}
			}
		}

		if (asJsonArray) {
			reader.endArray()
		} else {
			reader.endObject()
		}

		// если у нас есть все значения для конструктора (все значения присутствуют в исходном json объекте)
		// или у нас нет синтетического конструктора, то делаем вызов "обычного" конструктора
		if (syntheticFactory == null || presentValues.all { it }) {
			for ((i, tuple) in types.withIndex()) {
				val (field) = tuple

				if (readValues[i] == null) {
					if (!tuple.isMarkedNullable) {
						throw JsonSyntaxException("Field ${field.name} of ${bound.qualifiedName} does not accept nulls")
					}

					if (!regularFactory.parameters[i].isOptional && !presentValues[i]) {
						throw JsonSyntaxException("Field ${field.name} of ${bound.qualifiedName} must be defined (even just as null)")
					}
				}
			}

			@Suppress("unchecked_cast")
			return regularFactory.call(*readValues as Array<out Any>)
		// иначе - в бой вступает синтетический конструктор, с флагами "значения по умолчанию"
		} else {
			// количество bitflag'ов значений по умолчанию
			val ints = if (presentValues.size % 31 == 0) presentValues.size / 31 else presentValues.size / 31 + 1
			// максимум считанных значений + bitflag аргументы значений по умолчанию + null типа DefaultConstructorMarker
			val copied = readValues.copyOf(readValues.size + ints + 1)
			var intIndex = readValues.size
			var target = 0
			var targetMove = 0

			for (bool in presentValues) {
				if (!bool) {
					target = target.or(1.shl(targetMove))
				}

				targetMove++

				if (targetMove >= 32) {
					copied[intIndex++] = target
					target = 0
					targetMove = 0
				}
			}

			if (targetMove != 0) {
				copied[intIndex] = target
			}

			val syntheticPrimitives = syntheticPrimitives!!

			for ((i, tuple) in types.withIndex()) {
				if (copied[i] != null) {
					continue
				}

				val (field) = tuple
				val param = regularFactory.parameters[i]

				if (!param.isOptional && !presentValues[i]) {
					throw JsonSyntaxException("Field ${field.name} of ${bound.qualifiedName} is missing")
				}

				if (tuple.isMarkedNullable) {
					continue
				}

				if (param.isOptional && !presentValues[i]) {
					copied[i] = syntheticPrimitives[i]
					continue
				}

				throw JsonSyntaxException("Field \"${field.name}\" of ${bound.qualifiedName} does not accept nulls near ${reader.path}")
			}

			return syntheticFactory.newInstance(*copied)
		}
	}

	/**
	 * Позволяет построить класс [FactoryAdapter] на основе заданных параметров
	 */
	class Builder<T : Any>(val clazz: KClass<T>) : TypeAdapterFactory {
		constructor(clazz: KClass<T>, vararg fields: KProperty1<T, *>) : this(clazz) {
			for (field in fields) {
				auto(field)
			}
		}

		private var storesJson = false
		private var logMisses = true
		private val types = ArrayList<IResolvableProperty<T, *>>()
		private var stringTransformer: ((String) -> T)? = null
		var stringInterner: Interner<String> = Interner { it }

		fun stringInterner(interner: Interner<String>): Builder<T> {
			this.stringInterner = interner
			return this
		}

		fun ifString(transformer: (String) -> T): Builder<T> {
			stringTransformer = transformer
			return this
		}

		override fun <T : Any?> create(gson: Gson, type: TypeToken<T>): TypeAdapter<T>? {
			if (type.rawType == clazz.java) {
				return build(gson) as TypeAdapter<T>
			}

			return null
		}

		/**
		 * Собирает этот [FactoryAdapter] с указанным GSON объектом
		 */
		fun build(gson: Gson): TypeAdapter<T> {
			check(!asList || types.none { it.isFlat }) { "Can't have both flat properties and json data array layout" }

			return FactoryAdapter(
				bound = clazz,
				types = ImmutableList.copyOf(types.map { it.resolve(gson) }),
				asJsonArray = asList,
				storesJson = storesJson,
				logMisses = logMisses,
				stringInterner = stringInterner,
			).let {
				if (stringTransformer != null)
					it.ifString(stringTransformer!!)
				else
					it
			}
		}

		/**
		 * Собирает этот [FactoryAdapter] без GSON объекта
		 *
		 * Не рекомендуется использовать, лучше всего использовать как [TypeAdapterFactory]
		 *
		 * Несмотря на @Deprecated, данный вариант метода удалён не будет
		 */
		@Deprecated("Используйте как TypeAdapterFactory")
		fun build(): TypeAdapter<T> {
			check(!asList || types.none { it.isFlat }) { "Can't have both flat properties and json data array layout" }

			return FactoryAdapter(
				bound = clazz,
				types = ImmutableList.copyOf(types.map { it.resolve(null) }),
				asJsonArray = asList,
				storesJson = storesJson,
				logMisses = logMisses,
				stringInterner = stringInterner,
			).let {
				if (stringTransformer != null)
					it.ifString(stringTransformer!!)
				else
					it
			}
		}

		/**
		 * Принимает ли класс *последним* аргументом JSON структуру
		 *
		 * На самом деле, JSON "заворачивается" в [ImmutableMap], или [ImmutableList] если указано [asList]/[inputAsList]
		 *
		 * Поэтому, конструктор класса ОБЯЗАН принимать [Map]/[ImmutableMap] или [List]/[ImmutableList] последним аргументом,
		 * иначе поиск конструктора завершится неудачей
		 */
		fun storesJson(flag: Boolean = true): Builder<T> {
			storesJson = flag
			return this
		}

		/**
		 * Логировать ли json значения которые нет в списке свойств
		 */
		fun logMisses(flag: Boolean = true): Builder<T> {
			logMisses = flag
			return this
		}

		/**
		 * Добавляет свойство с определённым [adapter]
		 */
		fun <V> add(field: KProperty1<T, V>, adapter: TypeAdapter<V>, isFlat: Boolean = false): Builder<T> {
			types.add(ResolvedProperty(field, adapter, isFlat = isFlat))
			return this
		}

		/**
		 * Автоматически определяет необходимый адаптер типа к свойству при сборке данного адаптера внутри Gson
		 *
		 * Можно указать [transform] для изменения определённого адаптера
		 */
		@Suppress("unchecked_cast")
		fun <V> auto(field: KProperty1<T, V>, isFlat: Boolean = false, transform: (TypeAdapter<V>) -> TypeAdapter<V> = { it }): Builder<T> {
			types.add(ResolvableProperty(field, isFlat = isFlat, transform = transform))
			return this
		}

		private var asList = false

		/**
		 * При выставлении данного флага в качестве исходной структуры будет приниматься Json объект:
		 *
		 * ```json
		 * {
		 * 		"thingname": "a",
		 * 		"count": 4,
		 * 		...
		 * }
		 * ```
		 *
		 * Данный режим используется по умолчанию
		 *
		 * @see inputAsList
		 */
		fun inputAsMap(): Builder<T> {
			asList = false
			return this
		}

		/**
		 * При выставлении данного флага в качестве исходной структуры будет приниматься Json массив:
		 *
		 * ```json
		 * ["a", 4, 100.2, true]
		 * ```
		 *
		 * @see inputAsMap
		 */
		fun inputAsList(): Builder<T> {
			asList = true
			return this
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}

	class Factory(val stringInterner: Interner<String> = Interner { it }) : TypeAdapterFactory {
		override fun <T : Any> create(gson: Gson, type: TypeToken<T>): TypeAdapter<T>? {
			val raw = type.rawType

			if (raw.isAnnotationPresent(JsonFactory::class.java)) {
				val first = raw.getAnnotationsByType(JsonFactory::class.java)
				require(first.size == 1) { "Multiple JsonFactory defined: ${first.joinToString(", ")}" }

				val bconfig = first[0] as JsonFactory
				val kclass = raw.kotlin as KClass<T>
				val builder = Builder(kclass)
				val properties = kclass.declaredMembers.filterIsInstance<KProperty1<T, *>>()

				if (bconfig.asList) {
					builder.inputAsList()
				}

				builder.storesJson(bconfig.storesJson)
				builder.logMisses(bconfig.logMisses)
				builder.stringInterner = stringInterner

				if (properties.isEmpty()) {
					throw IllegalArgumentException("${kclass.qualifiedName} has no valid members")
				}

				val foundConstructor = kclass.primaryConstructor ?: throw NoSuchElementException("Can't determine primary constructor for ${kclass.qualifiedName}")

				if (!bconfig.storesJson) {
					for (argument in foundConstructor.parameters) {
						val property = properties.first { it.name == argument.name && it.returnType.isSupertypeOf(argument.type) }
						val config = property.annotations.firstOrNull { it.annotationClass == JsonPropertyConfig::class } as JsonPropertyConfig?
						builder.auto(property, isFlat = config?.isFlat ?: false)
					}
				} else {
					val params = foundConstructor.parameters

					for (i in 0 until params.size - 1) {
						val argument = params[i]
						val property = properties.first { it.name == argument.name && it.returnType.isSupertypeOf(argument.type) }
						val config = property.annotations.firstOrNull { it.annotationClass == JsonPropertyConfig::class } as JsonPropertyConfig?
						builder.auto(property, isFlat = config?.isFlat ?: false)
					}
				}

				return builder.build(gson)
			}

			return null
		}
	}
}
