package ru.dbotthepony.kstarbound.io.json.builder

import com.google.gson.Gson
import com.google.gson.TypeAdapter
import com.google.gson.reflect.TypeToken
import kotlin.reflect.KMutableProperty1
import kotlin.reflect.KProperty1
import kotlin.reflect.KType
import kotlin.reflect.javaType

interface IProperty {
	val isFlat: Boolean
	val name: String
	val mustBePresent: Boolean?
}

interface IReferencedProperty<T : Any, V> : IProperty {
	val property: KProperty1<T, V>
	override val name: String get() = property.name
	override val mustBePresent: Boolean?
		get() = null
}

interface IResolvableProperty<T : Any, V> : IReferencedProperty<T, V> {
	fun resolve(gson: Gson?): IResolvedProperty<T, V>
}

interface IResolvedProperty<T : Any, V> : IReferencedProperty<T, V> {
	val type: KType
	val adapter: TypeAdapter<V>
	val isMarkedNullable: Boolean

	operator fun component1() = property
	operator fun component2() = adapter
}

class ResolvedProperty<T : Any, V>(
	override val property: KProperty1<T, V>,
	override val adapter: TypeAdapter<V>,
	override val isFlat: Boolean
) : IResolvableProperty<T, V>, IResolvedProperty<T, V> {
	override val type: KType = property.returnType
	override val isMarkedNullable: Boolean = type.isMarkedNullable

	override fun resolve(gson: Gson?): IResolvedProperty<T, V> {
		return this
	}
}

class ResolvableProperty<T : Any, V>(
	override val property: KProperty1<T, V>,
	override val isFlat: Boolean,
	val transform: (TypeAdapter<V>) -> TypeAdapter<V> = { it }
) : IResolvableProperty<T, V> {
	@OptIn(ExperimentalStdlibApi::class)
	override fun resolve(gson: Gson?): IResolvedProperty<T, V> {
		gson ?: throw NullPointerException("Can not resolve without Gson present")

		return ResolvedProperty(
			property = property,
			adapter = transform(gson.getAdapter(TypeToken.get(property.returnType.javaType)) as TypeAdapter<V>),
			isFlat = isFlat
		)
	}
}

interface IReferencedMutableProperty<T : Any, V> : IProperty {
	val property: KMutableProperty1<T, V>
	override val name: String get() = property.name
}

interface IResolvableMutableProperty<T : Any, V> : IReferencedMutableProperty<T, V> {
	fun resolve(gson: Gson?): IResolvedMutableProperty<T, V>
}

interface IResolvedMutableProperty<T : Any, V> : IResolvableMutableProperty<T, V> {
	val type: KType
	val adapter: TypeAdapter<V>
	val isMarkedNullable: Boolean

	operator fun component1() = property
	operator fun component2() = adapter

	@Suppress("unchecked_cast")
	fun set(receiver: T, value: Any?) {
		property.set(receiver, value as V)
	}
}

class ResolvedMutableProperty<T : Any, V>(
	override val property: KMutableProperty1<T, V>,
	override val adapter: TypeAdapter<V>,
	override val isFlat: Boolean,
	override val mustBePresent: Boolean?
) : IResolvableMutableProperty<T, V>, IResolvedMutableProperty<T, V> {
	override val type: KType = property.returnType
	override val isMarkedNullable: Boolean = type.isMarkedNullable

	override fun resolve(gson: Gson?): IResolvedMutableProperty<T, V> {
		return this
	}
}

class ResolvableMutableProperty<T : Any, V>(
	override val property: KMutableProperty1<T, V>,
	override val isFlat: Boolean,
	val transform: (TypeAdapter<V>) -> TypeAdapter<V> = { it },
	override val mustBePresent: Boolean?
) : IResolvableMutableProperty<T, V> {
	@OptIn(ExperimentalStdlibApi::class)
	override fun resolve(gson: Gson?): IResolvedMutableProperty<T, V> {
		gson ?: throw NullPointerException("Can not resolve without Gson present")

		return ResolvedMutableProperty(
			property = property,
			adapter = transform(gson.getAdapter(TypeToken.get(property.returnType.javaType)) as TypeAdapter<V>),
			isFlat = isFlat,
			mustBePresent = mustBePresent,
		)
	}
}
