package ru.dbotthepony.kstarbound.lua

import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonNull
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.kenai.jffi.CallContext
import com.kenai.jffi.CallingConvention
import com.kenai.jffi.Closure
import com.kenai.jffi.ClosureManager
import com.kenai.jffi.Type
import com.sun.jna.Native
import com.sun.jna.ptr.LongByReference
import jnr.ffi.Memory
import jnr.ffi.NativeType
import jnr.ffi.Pointer
import org.lwjgl.system.MemoryUtil
import java.io.Closeable
import java.io.PrintWriter
import java.io.StringWriter
import java.nio.ByteBuffer
import java.nio.ByteOrder

private fun stringToBuffer(str: String): ByteBuffer {
	val bytes = str.toByteArray(charset = Charsets.UTF_8)
	val buf = ByteBuffer.allocateDirect(bytes.size)
	buf.order(ByteOrder.nativeOrder())
	bytes.forEach(buf::put)
	buf.position(0)

	return buf
}

class LuaState : Closeable {
	private val pointer = LuaJNR.INSTANCE.luaL_newstate() ?: throw OutOfMemoryError("Unable to allocate new LuaState")

	private var destroyed = false
	private val panicHandler = object : LuaJNR.lua_CFunction {
		override fun invoke(luaState: Pointer): Int {
			println("$this panicked!")
			return 0
		}
	}

	init {
		//LuaJNR.INSTANCE.lua_atpanic(pointer, panicHandler)

		LuaJNR.INSTANCE.luaopen_base(pointer)
		storeGlobal("_G")
		LuaJNR.INSTANCE.luaopen_package(pointer)
		storeGlobal("package")
		LuaJNR.INSTANCE.luaopen_table(pointer)
		storeGlobal("table")
		LuaJNR.INSTANCE.luaopen_coroutine(pointer)
		storeGlobal("coroutine")
		LuaJNR.INSTANCE.luaopen_string(pointer)
		storeGlobal("string")
		LuaJNR.INSTANCE.luaopen_math(pointer)
		storeGlobal("math")
		LuaJNR.INSTANCE.luaopen_utf8(pointer)
		storeGlobal("utf8")
		LuaJNR.INSTANCE.luaopen_debug(pointer)
		storeGlobal("debug")
	}

	fun pushClosure(lambda: (state: LuaState) -> Unit) {
		LuaJNI.lua_pushcclosure(pointer.address()) lazy@{
			try {
				lambda.invoke(this@LuaState)
			} catch (err: Throwable) {
				val builder = StringWriter()
				val printWriter = PrintWriter(builder)
				err.printStackTrace(printWriter)
				push(builder.toString())
				return@lazy 1
			}

			return@lazy 0
		}
	}

	fun checkStack(minAmount: Int): Boolean {
		return LuaJNR.INSTANCE.lua_checkstack(pointer, minAmount) > 0
	}

	fun absoluteIndex(index: Int): Int {
		return LuaJNR.INSTANCE.lua_absindex(pointer, index)
	}

	fun load(code: String, chunkName: String = "main chunk") {
		val buf = stringToBuffer(code)

		val closure = ClosureManager.getInstance().newClosure(
			object : Closure {
				override fun invoke(buffer: Closure.Buffer) {
					val remainingSize = LuaJNR.RUNTIME.memoryManager.newPointer(buffer.getAddress(2))

					if (buf.remaining() == 0) {
						remainingSize.putLong(0L, 0L)
						buffer.setAddressReturn(0L)
						return
					}

					remainingSize.putLongLong(0L, buf.remaining().toLong())
					val p = MemoryUtil.memAddress(buf)
					buf.position(buf.remaining())
					buffer.setAddressReturn(p)
				}
			},

			CallContext.getCallContext(Type.POINTER, arrayOf(Type.POINTER, Type.ULONG_LONG, Type.POINTER), CallingConvention.DEFAULT, false)
		)

		throwLoadError(LuaJNR.INSTANCE.lua_load(pointer, closure.address, 0L, chunkName, "t"))
		closure.dispose()
	}

	fun pcall(numArgs: Int = 0, numResults: Int = 0): Int {
		val status = LuaJNR.INSTANCE.lua_pcallk(pointer, numArgs, numResults, 0, 0L, 0L)

		if (status == LUA_ERRRUN) {
			throw LuaRuntimeException(getString())
		}

		return status
	}

	fun getString(index: Int = -1, limit: Long = 4096): String? {
		val len = Memory.allocateDirect(LuaJNR.RUNTIME, NativeType.SLONGLONG)
		val p = LuaJNR.INSTANCE.lua_tolstring(pointer, absoluteIndex(index), len) ?: return null

		if (len.getLong(0L) == 0L) {
			return ""
		}

		if (len.getLong(0L) >= limit) {
			throw IllegalStateException("Unreasonably long Lua string: ${len.getLong(0L)}")
		}

		val readBytes = ByteArray(len.getLong(0L).toInt())
		p.get(0L, readBytes, 0, readBytes.size)
		return readBytes.toString(charset = Charsets.UTF_8)
	}

	override fun close() {
		if (destroyed) {
			throw IllegalStateException("Already destroyed")
		}

		LuaJNR.INSTANCE.lua_close(pointer)
		destroyed = true
	}

	val stackTop: Int get() {
		val value = LuaJNR.INSTANCE.lua_gettop(pointer)
		check(value >= 0) { "Invalid stack top $value" }
		return value
	}

	fun storeGlobal(name: String) {
		LuaJNR.INSTANCE.lua_setglobal(pointer, name)
	}

	fun loadGlobal(name: String) {
		LuaJNR.INSTANCE.lua_getglobal(pointer, name)
	}

	fun push() {
		LuaJNR.INSTANCE.lua_pushnil(pointer)
	}

	fun push(value: Int) {
		LuaJNR.INSTANCE.lua_pushinteger(pointer, value.toLong())
	}

	fun push(value: Long) {
		LuaJNR.INSTANCE.lua_pushinteger(pointer, value)
	}

	fun push(value: String) {
		val bytes = value.toByteArray(Charsets.UTF_8)

		val block = LuaJNR.RUNTIME.memoryManager.allocateDirect(bytes.size) ?: throw OutOfMemoryError("Unable to allocate ${bytes.size} bytes on heap")

		//try {
			block.put(0L, bytes, 0, bytes.size)
			LuaJNR.INSTANCE.lua_pushlstring(pointer, block, bytes.size.toLong())
		//} finally {
		//	Memory.allocate()
		//}
	}

	fun push(value: JsonElement) {
		when (value) {
			JsonNull.INSTANCE -> {
				LuaJNR.INSTANCE.lua_pushnil(pointer)
			}

			is JsonPrimitive -> {
				if (value.isNumber) {
					val num = value.asNumber

					when (num) {
						is Int, is Long -> LuaJNR.INSTANCE.lua_pushinteger(pointer, num.toLong())
						else -> LuaJNR.INSTANCE.lua_pushnumber(pointer, num.toDouble())
					}
				} else if (value.isString) {
					push(value.asString)
				} else if (value.isBoolean) {
					LuaJNR.INSTANCE.lua_pushboolean(pointer, if (value.asBoolean) 1 else 0)
				} else {
					throw IllegalArgumentException(value.toString())
				}
			}

			is JsonArray -> {
				LuaJNR.INSTANCE.lua_createtable(pointer, value.size(), 0)
				val index = stackTop

				for ((i, v) in value.withIndex()) {
					LuaJNR.INSTANCE.lua_pushinteger(pointer, i.toLong() + 1L)
					push(v)

					LuaJNR.INSTANCE.lua_settable(pointer, index)
				}
			}

			is JsonObject -> {
				LuaJNR.INSTANCE.lua_createtable(pointer, 0, value.size())
				val index = stackTop

				for ((k, v) in value.entrySet()) {
					push(k)
					push(v)

					LuaJNR.INSTANCE.lua_settable(pointer, index)
				}
			}

			else -> {
				throw IllegalArgumentException(value.toString())
			}
		}
	}
}
